"""Implement functions which are used in the code decompiled by IDA."""

import sys
from typing import Type, TypeVar

from ..types import (IntVar, IntType, Int8, Int16, Int32, Int64, UInt8, UInt16,
                     UInt32, UInt64)
from ..utils import get_type

if sys.version_info >= (3, 8):
    from typing import Literal
else:
    from typing_extensions import Literal

T = TypeVar('T', Int8, Int16, Int32, Int64, UInt8, UInt16, UInt32, UInt64)

BIG_ENDIAN: Literal['big'] = 'big'
LITTLE_ENDIAN: Literal['little'] = 'little'

# Default use little endian.
BYTE_ORDER: Literal['big', 'little'] = LITTLE_ENDIAN

# Refer to defs.h of IDA.


def last_ind(x: IntVar, part_type: IntType) -> int:
    """Implementation of `LAST_IND`."""
    return x.get_size() // part_type.get_size() - 1


def low_ind(x: IntVar, part_type: IntType) -> int:
    """Implementation of `LOW_IND`."""
    return last_ind(x, part_type) if BYTE_ORDER == BIG_ENDIAN else 0


def high_ind(x: IntVar, part_type: IntType) -> int:
    """Implementation of `HIGH_IND`."""
    return 0 if BYTE_ORDER == BIG_ENDIAN else last_ind(x, part_type)


def offset_n(x: IntVar, n: int, t: Type[T]) -> T:
    """Get the nth member of the data."""
    size = t.get_size()
    data = x.to_bytes(byteorder=BYTE_ORDER)
    return t.from_bytes(data[n * size:(n + 1) * size], byteorder=BYTE_ORDER)


def byten(x: IntVar, n: int) -> UInt8:
    """Implementation of `BYTEn`."""
    return offset_n(x, n, UInt8)


def wordn(x: IntVar, n: int) -> UInt16:
    """Implementation of `WORDn`."""
    return offset_n(x, n, UInt16)


def dwordn(x: IntVar, n: int) -> UInt32:
    """Implementation of `DWORDn`."""
    return offset_n(x, n, UInt32)


def lobyte(x: IntVar) -> UInt8:
    """Implementation of `LOBYTE`."""
    return byten(x, low_ind(x, UInt8))


def loword(x: IntVar) -> UInt16:
    """Implementation of `LOWORD`."""
    return wordn(x, low_ind(x, UInt16))


def lodword(x: IntVar) -> UInt32:
    """Implementation of `LODWORD`."""
    return dwordn(x, low_ind(x, UInt32))


def hibyte(x: IntVar) -> UInt8:
    """Implementation of `HIBYTE`."""
    return byten(x, high_ind(x, UInt8))


def hiword(x: IntVar) -> UInt16:
    """Implementation of `HIWORD`."""
    return wordn(x, high_ind(x, UInt16))


def hidword(x: IntVar) -> UInt32:
    """Implementation of `HIDWORD`."""
    return dwordn(x, high_ind(x, UInt32))


def byte1(x: IntVar) -> UInt8:
    """Implementation of `BYTE1`."""
    return byten(x, 1)


def byte2(x: IntVar) -> UInt8:
    """Implementation of `BYTE2`."""
    return byten(x, 2)


def byte3(x: IntVar) -> UInt8:
    """Implementation of `BYTE3`."""
    return byten(x, 3)


def byte4(x: IntVar) -> UInt8:
    """Implementation of `BYTE4`."""
    return byten(x, 4)


def byte5(x: IntVar) -> UInt8:
    """Implementation of `BYTE5`."""
    return byten(x, 5)


def byte6(x: IntVar) -> UInt8:
    """Implementation of `BYTE6`."""
    return byten(x, 6)


def byte7(x: IntVar) -> UInt8:
    """Implementation of `BYTE7`."""
    return byten(x, 7)


def byte8(x: IntVar) -> UInt8:
    """Implementation of `BYTE8`."""
    return byten(x, 8)


def byte9(x: IntVar) -> UInt8:
    """Implementation of `BYTE9`."""
    return byten(x, 9)


def byte10(x: IntVar) -> UInt8:
    """Implementation of `BYTE10`."""
    return byten(x, 10)


def byte11(x: IntVar) -> UInt8:
    """Implementation of `BYTE11`."""
    return byten(x, 11)


def byte12(x: IntVar) -> UInt8:
    """Implementation of `BYTE12`."""
    return byten(x, 12)


def byte13(x: IntVar) -> UInt8:
    """Implementation of `BYTE13`."""
    return byten(x, 13)


def byte14(x: IntVar) -> UInt8:
    """Implementation of `BYTE14`."""
    return byten(x, 14)


def byte15(x: IntVar) -> UInt8:
    """Implementation of `BYTE15`."""
    return byten(x, 15)


def word1(x: IntVar) -> UInt16:
    """Implementation of `WORD1`."""
    return wordn(x, 1)


def word2(x: IntVar) -> UInt16:
    """Implementation of `WORD2`."""
    return wordn(x, 2)


def word3(x: IntVar) -> UInt16:
    """Implementation of `WORD3`."""
    return wordn(x, 3)


def word4(x: IntVar) -> UInt16:
    """Implementation of `WORD4`."""
    return wordn(x, 4)


def word5(x: IntVar) -> UInt16:
    """Implementation of `WORD5`."""
    return wordn(x, 5)


def word6(x: IntVar) -> UInt16:
    """Implementation of `WORD6`."""
    return wordn(x, 6)


def word7(x: IntVar) -> UInt16:
    """Implementation of `WORD7`."""
    return wordn(x, 7)


def dword1(x: IntVar) -> UInt32:
    """Implementation of `DWORD1`."""
    return dwordn(x, 1)


def dword2(x: IntVar) -> UInt32:
    """Implementation of `DWORD2`."""
    return dwordn(x, 2)


def dword3(x: IntVar) -> UInt32:
    """Implementation of `DWORD3`."""
    return dwordn(x, 3)


def sbyten(x: IntVar, n: int) -> Int8:
    """Implementation of `SBYTEn`."""
    return offset_n(x, n, Int8)


def swordn(x: IntVar, n: int) -> Int16:
    """Implementation of `SWORDn`."""
    return offset_n(x, n, Int16)


def sdwordn(x: IntVar, n: int) -> Int32:
    """Implementation of `SDWORDn`."""
    return offset_n(x, n, Int32)


def slobyte(x: IntVar) -> Int8:
    """Implementation of `SLOBYTE`."""
    return sbyten(x, low_ind(x, Int8))


def sloword(x: IntVar) -> Int16:
    """Implementation of `SLOWORD`."""
    return swordn(x, low_ind(x, Int16))


def slodword(x: IntVar) -> Int32:
    """Implementation of `SLODWORD`."""
    return sdwordn(x, low_ind(x, Int32))


def shibyte(x: IntVar) -> Int8:
    """Implementation of `SHIBYTE`."""
    return sbyten(x, high_ind(x, Int8))


def shiword(x: IntVar) -> Int16:
    """Implementation of `SHIWORD`."""
    return swordn(x, high_ind(x, Int16))


def shidword(x: IntVar) -> Int32:
    """Implementation of `SHIDWORD`."""
    return sdwordn(x, high_ind(x, Int32))


def sbyte1(x: IntVar) -> Int8:
    """Implementation of `SBYTE1`."""
    return sbyten(x, 1)


def sbyte2(x: IntVar) -> Int8:
    """Implementation of `SBYTE2`."""
    return sbyten(x, 2)


def sbyte3(x: IntVar) -> Int8:
    """Implementation of `SBYTE3`."""
    return sbyten(x, 3)


def sbyte4(x: IntVar) -> Int8:
    """Implementation of `SBYTE4`."""
    return sbyten(x, 4)


def sbyte5(x: IntVar) -> Int8:
    """Implementation of `SBYTE5`."""
    return sbyten(x, 5)


def sbyte6(x: IntVar) -> Int8:
    """Implementation of `SBYTE6`."""
    return sbyten(x, 6)


def sbyte7(x: IntVar) -> Int8:
    """Implementation of `SBYTE7`."""
    return sbyten(x, 7)


def sbyte8(x: IntVar) -> Int8:
    """Implementation of `SBYTE8`."""
    return sbyten(x, 8)


def sbyte9(x: IntVar) -> Int8:
    """Implementation of `SBYTE9`."""
    return sbyten(x, 9)


def sbyte10(x: IntVar) -> Int8:
    """Implementation of `SBYTE10`."""
    return sbyten(x, 10)


def sbyte11(x: IntVar) -> Int8:
    """Implementation of `SBYTE11`."""
    return sbyten(x, 11)


def sbyte12(x: IntVar) -> Int8:
    """Implementation of `SBYTE12`."""
    return sbyten(x, 12)


def sbyte13(x: IntVar) -> Int8:
    """Implementation of `SBYTE13`."""
    return sbyten(x, 13)


def sbyte14(x: IntVar) -> Int8:
    """Implementation of `SBYTE14`."""
    return sbyten(x, 14)


def sbyte15(x: IntVar) -> Int8:
    """Implementation of `SBYTE15`."""
    return sbyten(x, 15)


def sword1(x: IntVar) -> Int16:
    """Implementation of `SWORD1`."""
    return swordn(x, 1)


def sword2(x: IntVar) -> Int16:
    """Implementation of `SWORD2`."""
    return swordn(x, 2)


def sword3(x: IntVar) -> Int16:
    """Implementation of `SWORD3`."""
    return swordn(x, 3)


def sword4(x: IntVar) -> Int16:
    """Implementation of `SWORD4`."""
    return swordn(x, 4)


def sword5(x: IntVar) -> Int16:
    """Implementation of `SWORD5`."""
    return swordn(x, 5)


def sword6(x: IntVar) -> Int16:
    """Implementation of `SWORD6`."""
    return swordn(x, 6)


def sword7(x: IntVar) -> Int16:
    """Implementation of `SWORD7`."""
    return swordn(x, 7)


def sdword1(x: IntVar) -> Int32:
    """Implementation of `SDWORD1`."""
    return sdwordn(x, 1)


def sdword2(x: IntVar) -> Int32:
    """Implementation of `SDWORD2`."""
    return sdwordn(x, 2)


def sdword3(x: IntVar) -> Int32:
    """Implementation of `SDWORD3`."""
    return sdwordn(x, 3)


def pair(high: IntVar, low: IntVar) -> IntVar:
    """Implementation of `__PAIR__`."""
    size = high.get_size()
    signed = high.get_signed()
    int_type = get_type(size=size * 2, signed=signed)
    return int_type(high) << size * 8 | type(high)(low)


def rol(value: T, count: int) -> T:
    """Implementation of `__ROL__`."""
    data_type = type(value)
    nbits = value.get_size() * 8

    if count > 0:
        count %= nbits
        high = value >> (nbits - count)
        if value.get_signed():
            high &= ~(data_type(-1) << count)
        value <<= count
        value |= high

    else:
        count = -count % nbits
        low = value << (nbits - count)
        value >>= count
        value |= low

    return value


def rol1(value: UInt8, count: int) -> UInt8:
    """Implementation of `__ROL1__`."""
    return rol(value, count)


def rol2(value: UInt16, count: int) -> UInt16:
    """Implementation of `__ROL2__`."""
    return rol(value, count)


def rol4(value: UInt32, count: int) -> UInt32:
    """Implementation of `__ROL4__`."""
    return rol(value, count)


def rol8(value: UInt64, count: int) -> UInt64:
    """Implementation of `__ROL8__`."""
    return rol(value, count)


def ror1(value: UInt8, count: int) -> UInt8:
    """Implementation of `__ROR1__`."""
    return rol(value, -count)


def ror2(value: UInt16, count: int) -> UInt16:
    """Implementation of `__ROR2__`."""
    return rol(value, -count)


def ror4(value: UInt32, count: int) -> UInt32:
    """Implementation of `__ROR4__`."""
    return rol(value, -count)


def ror8(value: UInt64, count: int) -> UInt64:
    """Implementation of `__ROR8__`."""
    return rol(value, -count)


def mkcshl(value: IntVar, count: int) -> Int8:
    """Implementation of `__MKCSHL__`."""
    nbits = value.get_size() * 8
    count %= nbits
    return Int8((value >> (nbits - count)) & 1)


def mkcshr(value: IntVar, count: int) -> Int8:
    """Implementation of `__MKCSHR__`."""
    return Int8((value >> (count - 1)) & 1)


def sets(x: IntVar) -> Int8:
    """Implementation of `__SETS__`."""
    data_type = get_type(size=x.get_size(), signed=True)
    return Int8(data_type(x) < 0)


def ofsub(x: IntVar, y: IntVar) -> Int8:
    """Implementation of `__OFSUB__`."""
    if x.get_size() < y.get_size():
        x2 = x
        sx = sets(x2)
        return Int8((sx ^ sets(y)) & (sx ^ sets(x2 - y)))
    else:
        y2 = y
        sx = sets(x)
        return Int8((sx ^ sets(y2)) & (sx ^ sets(x - y2)))


def ofadd(x: IntVar, y: IntVar) -> Int8:
    """Implementation of `__OFADD__`."""
    if x.get_size() < y.get_size():
        x2 = x
        sx = sets(x2)
        return Int8(((1 ^ sx) ^ sets(y)) & (sx ^ sets(x2 + y)))
    else:
        y2 = y
        sx = sets(x)
        return Int8(((1 ^ sx) ^ sets(y2)) & (sx ^ sets(x + y2)))


def cfsub(x: IntVar, y: IntVar) -> Int8:
    """Implementation of `__CFSUB__`."""
    size = max(x.get_size(), y.get_size())
    data_type = get_type(size=size, signed=False)
    return Int8(data_type(x) < data_type(y))


def cfadd(x: IntVar, y: IntVar) -> Int8:
    """Implementation of `__CFADD__`."""
    size = max(x.get_size(), y.get_size())
    data_type = get_type(size=size, signed=False)
    return Int8(data_type(x) > data_type(x + y))


# Refer to https://gcc.gnu.org/onlinedocs/gcc/Other-Builtins.html.


def swap_bytes(value: T) -> T:
    """Reverse bytes of the value with the order."""
    data = value.to_bytes(byteorder=BYTE_ORDER)
    return value.from_bytes(data[::-1], byteorder=BYTE_ORDER)


def bswap16(value: UInt16) -> UInt16:
    """Implementation of `bswap16`."""
    return swap_bytes(value)


def bswap32(value: UInt32) -> UInt32:
    """Implementation of `bswap32`."""
    return swap_bytes(value)


def bswap64(value: UInt64) -> UInt64:
    """Implementation of `bswap64`."""
    return swap_bytes(value)


def clz(x: IntVar) -> Int8:
    """Implementation of `__clz`."""
    return Int8(x.get_size() * 8 - len(bin(x)[2:]))



from typing import List, Union

from .exceptions import InvalidOperationError
from .types import ByteOrder, IntBase, IntType, IntVar, UInt8
from .utils import get_type


class VirtualPointer:
    """Provide virtual pointer operation on bytearray."""

    def __init__(self,
                 source: bytearray,
                 data_type: Union[IntType, str],
                 byteorder: ByteOrder = 'little',
                 offset: int = 0):
        self.source = source
        self.byteorder = byteorder
        self.offset = offset

        self._data_type = None
        self.data_type = data_type

    def __add__(self, other):
        """Support addition."""
        return self.add(other)

    def __sub__(self, other):
        """Support subtraction."""
        return self.sub(other)

    @property
    def data_type(self):
        """Get data type."""
        return self._data_type

    @data_type.setter
    def data_type(self, value: Union[IntType, str]):
        """Set data type."""
        if isinstance(value, str):
            try:
                self._data_type = get_type(type_name=value)
            except NotImplementedError as e:
                raise InvalidOperationError('Unsupported type') from e
        elif issubclass(value, IntBase):
            self._data_type = value
        else:
            raise TypeError()

    def copy(self) -> "VirtualPointer":
        """Copy this object."""
        return self.__class__(source=self.source,
                              data_type=self.data_type,
                              byteorder=self.byteorder,
                              offset=self.offset)

    def add(self, num: int) -> "VirtualPointer":
        """Offset this pointer position."""
        obj = self.copy()
        obj.offset += num * self.data_type.get_size()
        return obj

    def sub(self, num: int) -> "VirtualPointer":
        """Reverse offset this pointer position."""
        return self.add(-num)

    def cast(self, data_type: Union[IntType, str]) -> "VirtualPointer":
        """Cast to the specified type."""
        obj = self.copy()
        obj.data_type = data_type
        return obj

    def read_bytes(self, size: int) -> bytes:
        """Read bytes from the source bytearray."""
        if self.offset + size > len(self.source):
            raise InvalidOperationError('Read out of range')
        return bytes(self.source[self.offset:self.offset + size])

    def write_bytes(self, data: Union[bytes, bytearray, List[int]]):
        """Write bytes into the source bytearray."""
        try:
            for i, v in enumerate(data):
                self.source[self.offset + i] = v
        except IndexError as e:
            raise InvalidOperationError('Write out of range') from e

    def read(self) -> IntVar:
        """Read an integer from the source bytearray."""
        data = self.read_bytes(self.data_type.get_size())
        return self.data_type.from_bytes(data, byteorder=self.byteorder)

    def write(self, value: Union[IntVar, int]):
        """Write an integer into the source bytearray."""
        data = self.data_type(value).to_bytes(byteorder=self.byteorder)
        self.write_bytes(data)


def vptr(
    source: bytearray,
    data_type: Union[IntType, str] = UInt8,
    byteorder: ByteOrder = 'little',
) -> VirtualPointer:
    """Shorthand of `VirtualPointer(source, data_type)`."""
    return VirtualPointer(source=source,
                          data_type=data_type,
                          byteorder=byteorder)



"""Implement functions which are used in the code decompiled by Ghidra."""

from typing import Type, TypeVar

from .ida import cfadd, ofsub, ofadd
from ..types import Int8, Int16, Int32, Int64, UInt8, UInt16, UInt32, UInt64
from ..utils import get_type

# Refer to https://github.com/NationalSecurityAgency/ghidra/blob/master/Ghidra
# /Features/Decompiler/src/main/help/help/topics/DecompilePlugin
# /DecompilerConcepts.html

InVar = TypeVar('InVar', Int8, Int16, Int32, Int64, UInt8, UInt16, UInt32,
                UInt64)
OutVar = TypeVar('OutVar', Int8, Int16, Int32, Int64, UInt8, UInt16, UInt32,
                 UInt64)


def truncate(x: InVar, c: int, to_type: Type[OutVar]) -> OutVar:
    """Truncation operation."""
    return to_type.from_bytes(x.to_bytes()[c:c + to_type.get_size()])


def sub21(x: UInt16, c: int) -> UInt8:
    """Implementation of `SUB21`."""
    return truncate(x, c, UInt8)


def sub41(x: UInt32, c: int) -> UInt8:
    """Implementation of `SUB41`."""
    return truncate(x, c, UInt8)


def sub42(x: UInt32, c: int) -> UInt16:
    """Implementation of `SUB42`."""
    return truncate(x, c, UInt16)


def sub81(x: UInt64, c: int) -> UInt8:
    """Implementation of `SUB81`."""
    return truncate(x, c, UInt8)


def sub82(x: UInt64, c: int) -> UInt16:
    """Implementation of `SUB82`."""
    return truncate(x, c, UInt16)


def sub84(x: UInt64, c: int) -> UInt32:
    """Implementation of `SUB84`."""
    return truncate(x, c, UInt32)


def zero_extend(x: InVar, from_type: Type[InVar],
                to_type: Type[OutVar]) -> OutVar:
    """Zero-extension operator."""
    return to_type.from_bytes(from_type(x).to_bytes())


def zext12(x: UInt8) -> UInt16:
    """Implementation of `ZEXT12`."""
    return zero_extend(x, UInt8, UInt16)


def zext14(x: UInt8) -> UInt32:
    """Implementation of `ZEXT14`."""
    return zero_extend(x, UInt8, UInt32)


def zext18(x: UInt8) -> UInt64:
    """Implementation of `ZEXT18`."""
    return zero_extend(x, UInt8, UInt64)


def zext24(x: UInt16) -> UInt32:
    """Implementation of `ZEXT14`."""
    return zero_extend(x, UInt16, UInt32)


def zext28(x: UInt16) -> UInt64:
    """Implementation of `ZEXT18`."""
    return zero_extend(x, UInt16, UInt64)


def zext48(x: UInt32) -> UInt64:
    """Implementation of `ZEXT48`."""
    return zero_extend(x, UInt32, UInt64)


def sign_extend(x: InVar, from_type: Type[InVar],
                to_type: Type[OutVar]) -> OutVar:
    """Sign-extension operator."""
    t1 = get_type(size=from_type.get_size(), signed=True)
    t2 = get_type(size=to_type.get_size(), signed=True)
    return to_type.from_bytes(t2.from_bytes(t1(x).to_bytes()).to_bytes())


def sext12(x: UInt8) -> UInt16:
    """Implementation of `SEXT12`."""
    return sign_extend(x, UInt8, UInt16)


def sext14(x: UInt8) -> UInt32:
    """Implementation of `SEXT14`."""
    return sign_extend(x, UInt8, UInt32)


def sext18(x: UInt8) -> UInt64:
    """Implementation of `SEXT18`."""
    return sign_extend(x, UInt8, UInt64)


def sext24(x: UInt16) -> UInt32:
    """Implementation of `SEXT14`."""
    return sign_extend(x, UInt16, UInt32)


def sext28(x: UInt16) -> UInt64:
    """Implementation of `SEXT18`."""
    return sign_extend(x, UInt16, UInt64)


def sext48(x: UInt32) -> UInt64:
    """Implementation of `SEXT48`."""
    return sign_extend(x, UInt32, UInt64)


def sborrow1(x: UInt8, y: UInt8) -> Int8:
    """Implementation of `SBORROW1`."""
    return ofsub(x, y)


def sborrow2(x: UInt16, y: UInt16) -> Int8:
    """Implementation of `SBORROW2`."""
    return ofsub(x, y)


def sborrow4(x: UInt32, y: UInt32) -> Int8:
    """Implementation of `SBORROW4`."""
    return ofsub(x, y)


def sborrow8(x: UInt64, y: UInt64) -> Int8:
    """Implementation of `SBORROW8`."""
    return ofsub(x, y)


def carry1(x: UInt8, y: UInt8) -> Int8:
    """Implementation of `CARRY1`."""
    return cfadd(x, y)


def carry2(x: UInt16, y: UInt16) -> Int8:
    """Implementation of `CARRY2`."""
    return cfadd(x, y)


def carry4(x: UInt32, y: UInt32) -> Int8:
    """Implementation of `CARRY4`."""
    return cfadd(x, y)


def carry8(x: UInt64, y: UInt64) -> Int8:
    """Implementation of `CARRY8`."""
    return cfadd(x, y)


def scarry1(x: Int8, y: Int8) -> Int8:
    """Implementation of `SCARRY1`."""
    return ofadd(x, y)


def scarry2(x: Int16, y: Int16) -> Int8:
    """Implementation of `SCARRY2`."""
    return ofadd(x, y)


def scarry4(x: Int32, y: Int32) -> Int8:
    """Implementation of `SCARRY4`."""
    return ofadd(x, y)


def scarry8(x: Int64, y: Int64) -> Int8:
    """Implementation of `SCARRY8`."""
    return ofadd(x, y)



import re
from typing import Optional, List

import numpy as np

from .types import (IntType, Int8, Int16, Int32, Int64, UInt8, UInt16, UInt32,
                    UInt64)


def get_type(size: Optional[int] = None,
             signed: Optional[bool] = None,
             type_name: Optional[str] = None) -> IntType:
    """Get type int with specified size and signed."""
    if type_name is not None:
        match = re.match(r'(u*)int(\d+)', type_name)
        if not match:
            raise NotImplementedError()

        nbits = int(match.group(2))
        if nbits % 8:
            raise NotImplementedError()

        signed = not bool(match.group(1))
        size = nbits // 8

    int_types: List[IntType] = [
        Int8, Int16, Int32, Int64, UInt8, UInt16, UInt32, UInt64
    ]

    for int_type in int_types:
        if int_type.get_size() == size and int_type.get_signed() == signed:
            return int_type

    raise NotImplementedError()


def disable_numpy_overflow_warnings():
    """Disable overflow warnings of numpy."""
    np.seterr(over='ignore')



import re
import sys
from functools import wraps
from typing import Union, Type, SupportsBytes, Iterable

import numpy as np

if sys.version_info >= (3, 8):
    from typing import Literal
else:
    from typing_extensions import Literal

__all__ = [
    'ByteOrder', 'IntVar', 'IntType', 'IntMeta', 'IntBase', 'Int8', 'Int16',
    'Int32', 'Int64', 'UInt8', 'UInt16', 'UInt32', 'UInt64', 'int8', 'int16',
    'int32', 'int64', 'uint8', 'uint16', 'uint32', 'uint64'
]

ByteOrder = Literal['little', 'big']

IntVar = Union['Int8', 'Int16', 'Int32', 'Int64', 'UInt8', 'UInt16', 'UInt32',
               'UInt64']

IntType = Type[Union['Int8', 'Int16', 'Int32', 'Int64', 'UInt8', 'UInt16',
                     'UInt32', 'UInt64']]


class IntMeta(type):
    """Meta class of integer type."""

    _OVERRIDE_OPERATIONS = [
        '__neg__', '__pos__', '__invert__', '__add__', '__radd__', '__iadd__',
        '__sub__', '__rsub__', '__isub__', '__mul__', '__rmul__', '__imul__',
        '__truediv__', '__rtruediv__', '__itruediv__', '__floordiv__',
        '__rfloordiv__', '__ifloordiv__', '__mod__', '__rmod__', '__imod__',
        '__and__', '__rand__', '__iand__', '__or__', '__ror__', '__ior__',
        '__xor__', '__rxor__', '__ixor__', '__lshift__', '__rlshift__',
        '__ilshift__', '__rshift__', '__rrshift__', '__irshift__'
    ]

    def __init__(cls, name, bases, attr_dict):
        super().__init__(name, bases, attr_dict)

        np_base = None
        for base in bases:
            if issubclass(base, np.integer):
                np_base = base
        if not np_base:
            raise TypeError('Invalid class')

        setattr(cls, '_np_base', np_base)

        size = int(re.match(r'(U*)Int(\d+)', cls.__name__).group(2)) // 8
        setattr(cls, '_data_size', size)

        for op in cls._OVERRIDE_OPERATIONS:
            if not hasattr(cls, op):
                continue
            setattr(cls, op, cls.operation_wrapper(getattr(cls, op)))

    @staticmethod
    def operation_wrapper(f):
        """Wrap operation method for overriding type conversion.

        If it is a unary operation, the result is still its own type.
        If the other operand is int, it will be converted to its own type.
        If the type of another operand is inconsistent with its own type,
        they are converted to larger size or unsigned type.
        """

        @wraps(f)
        def decorator(self, *args):
            data_type = type(self)
            operator = getattr(self._np_base, f.__name__)

            if args:
                other = args[0]

                if isinstance(other, int):
                    other = data_type(other)

                elif isinstance(other, np.integer):
                    if self.itemsize == other.itemsize:
                        data_type = type(other if isinstance(
                            self, np.signedinteger) else self)
                    else:
                        if self.itemsize < other.itemsize:
                            data_type = type(other)
                            operator = getattr(other, f.__name__)
                            self, other = other, self

                        else:
                            data_type = type(self)

                return data_type(operator(self, other))

            return data_type(f(self))

        return decorator


class IntBase:
    """Base class of integer type."""

    _data_size: int

    def __init__(self, *args, **kwargs):
        pass

    def __int__(self) -> int:
        pass

    @classmethod
    def get_size(cls) -> int:
        """Get size (bytes) of this type."""
        return cls._data_size

    @classmethod
    def get_signed(cls) -> bool:
        """Get signed of this type."""
        return issubclass(cls, np.signedinteger)

    @classmethod
    def from_bytes(cls,
                   data: Union[Iterable[int], SupportsBytes],
                   byteorder: ByteOrder = 'little'):
        """Return a value of this type from given bytes"""
        return cls(
            int.from_bytes(data, byteorder=byteorder, signed=cls.get_signed()))

    def to_bytes(self, byteorder: ByteOrder = 'little') -> bytes:
        """Covert this value to bytes."""
        return int(self).to_bytes(length=self.get_size(),
                                  byteorder=byteorder,
                                  signed=self.get_signed())


class Int8(np.int8, IntBase, metaclass=IntMeta):
    """Int8"""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def __add__(self, other):
        pass

    def __radd__(self, other):
        pass

    def __sub__(self, other):
        pass

    def __rsub__(self, other):
        pass

    def __mul__(self, other):
        pass

    def __rmul__(self, other):
        pass

    def __truediv__(self, other):
        pass

    def __rtruediv__(self, other):
        pass

    def __floordiv__(self, other):
        pass

    def __rfloordiv__(self, other):
        pass

    def __mod__(self, other):
        pass

    def __rmod__(self, other):
        pass

    def __and__(self, other):
        pass

    def __rand__(self, other):
        pass

    def __or__(self, other):
        pass

    def __ror__(self, other):
        pass

    def __xor__(self, other):
        pass

    def __rxor__(self, other):
        pass

    def __lshift__(self, other):
        pass

    def __rlshift__(self, other):
        pass

    def __rshift__(self, other):
        pass

    def __rrshift__(self, other):
        pass


class Int16(np.int16, IntBase, metaclass=IntMeta):
    """Int16"""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def __add__(self, other):
        pass

    def __radd__(self, other):
        pass

    def __sub__(self, other):
        pass

    def __rsub__(self, other):
        pass

    def __mul__(self, other):
        pass

    def __rmul__(self, other):
        pass

    def __truediv__(self, other):
        pass

    def __rtruediv__(self, other):
        pass

    def __floordiv__(self, other):
        pass

    def __rfloordiv__(self, other):
        pass

    def __mod__(self, other):
        pass

    def __rmod__(self, other):
        pass

    def __and__(self, other):
        pass

    def __rand__(self, other):
        pass

    def __or__(self, other):
        pass

    def __ror__(self, other):
        pass

    def __xor__(self, other):
        pass

    def __rxor__(self, other):
        pass

    def __lshift__(self, other):
        pass

    def __rlshift__(self, other):
        pass

    def __rshift__(self, other):
        pass

    def __rrshift__(self, other):
        pass


class Int32(np.int32, IntBase, metaclass=IntMeta):
    """Int32"""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def __add__(self, other):
        pass

    def __radd__(self, other):
        pass

    def __sub__(self, other):
        pass

    def __rsub__(self, other):
        pass

    def __mul__(self, other):
        pass

    def __rmul__(self, other):
        pass

    def __truediv__(self, other):
        pass

    def __rtruediv__(self, other):
        pass

    def __floordiv__(self, other):
        pass

    def __rfloordiv__(self, other):
        pass

    def __mod__(self, other):
        pass

    def __rmod__(self, other):
        pass

    def __and__(self, other):
        pass

    def __rand__(self, other):
        pass

    def __or__(self, other):
        pass

    def __ror__(self, other):
        pass

    def __xor__(self, other):
        pass

    def __rxor__(self, other):
        pass

    def __lshift__(self, other):
        pass

    def __rlshift__(self, other):
        pass

    def __rshift__(self, other):
        pass

    def __rrshift__(self, other):
        pass


class Int64(np.int64, IntBase, metaclass=IntMeta):
    """Int64"""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def __add__(self, other):
        pass

    def __radd__(self, other):
        pass

    def __sub__(self, other):
        pass

    def __rsub__(self, other):
        pass

    def __mul__(self, other):
        pass

    def __rmul__(self, other):
        pass

    def __truediv__(self, other):
        pass

    def __rtruediv__(self, other):
        pass

    def __floordiv__(self, other):
        pass

    def __rfloordiv__(self, other):
        pass

    def __mod__(self, other):
        pass

    def __rmod__(self, other):
        pass

    def __and__(self, other):
        pass

    def __rand__(self, other):
        pass

    def __or__(self, other):
        pass

    def __ror__(self, other):
        pass

    def __xor__(self, other):
        pass

    def __rxor__(self, other):
        pass

    def __lshift__(self, other):
        pass

    def __rlshift__(self, other):
        pass

    def __rshift__(self, other):
        pass

    def __rrshift__(self, other):
        pass


class UInt8(np.uint8, IntBase, metaclass=IntMeta):
    """UInt8"""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def __add__(self, other):
        pass

    def __radd__(self, other):
        pass

    def __sub__(self, other):
        pass

    def __rsub__(self, other):
        pass

    def __mul__(self, other):
        pass

    def __rmul__(self, other):
        pass

    def __truediv__(self, other):
        pass

    def __rtruediv__(self, other):
        pass

    def __floordiv__(self, other):
        pass

    def __rfloordiv__(self, other):
        pass

    def __mod__(self, other):
        pass

    def __rmod__(self, other):
        pass

    def __and__(self, other):
        pass

    def __rand__(self, other):
        pass

    def __or__(self, other):
        pass

    def __ror__(self, other):
        pass

    def __xor__(self, other):
        pass

    def __rxor__(self, other):
        pass

    def __lshift__(self, other):
        pass

    def __rlshift__(self, other):
        pass

    def __rshift__(self, other):
        pass

    def __rrshift__(self, other):
        pass


class UInt16(np.uint16, IntBase, metaclass=IntMeta):
    """UInt16"""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def __add__(self, other):
        pass

    def __radd__(self, other):
        pass

    def __sub__(self, other):
        pass

    def __rsub__(self, other):
        pass

    def __mul__(self, other):
        pass

    def __rmul__(self, other):
        pass

    def __truediv__(self, other):
        pass

    def __rtruediv__(self, other):
        pass

    def __floordiv__(self, other):
        pass

    def __rfloordiv__(self, other):
        pass

    def __mod__(self, other):
        pass

    def __rmod__(self, other):
        pass

    def __and__(self, other):
        pass

    def __rand__(self, other):
        pass

    def __or__(self, other):
        pass

    def __ror__(self, other):
        pass

    def __xor__(self, other):
        pass

    def __rxor__(self, other):
        pass

    def __lshift__(self, other):
        pass

    def __rlshift__(self, other):
        pass

    def __rshift__(self, other):
        pass

    def __rrshift__(self, other):
        pass


class UInt32(np.uint32, IntBase, metaclass=IntMeta):
    """UInt32"""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def __add__(self, other):
        pass

    def __radd__(self, other):
        pass

    def __sub__(self, other):
        pass

    def __rsub__(self, other):
        pass

    def __mul__(self, other):
        pass

    def __rmul__(self, other):
        pass

    def __truediv__(self, other):
        pass

    def __rtruediv__(self, other):
        pass

    def __floordiv__(self, other):
        pass

    def __rfloordiv__(self, other):
        pass

    def __mod__(self, other):
        pass

    def __rmod__(self, other):
        pass

    def __and__(self, other):
        pass

    def __rand__(self, other):
        pass

    def __or__(self, other):
        pass

    def __ror__(self, other):
        pass

    def __xor__(self, other):
        pass

    def __rxor__(self, other):
        pass

    def __lshift__(self, other):
        pass

    def __rlshift__(self, other):
        pass

    def __rshift__(self, other):
        pass

    def __rrshift__(self, other):
        pass


class UInt64(np.uint64, IntBase, metaclass=IntMeta):
    """UInt64"""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def __add__(self, other):
        pass

    def __radd__(self, other):
        pass

    def __sub__(self, other):
        pass

    def __rsub__(self, other):
        pass

    def __mul__(self, other):
        pass

    def __rmul__(self, other):
        pass

    def __truediv__(self, other):
        pass

    def __rtruediv__(self, other):
        pass

    def __floordiv__(self, other):
        pass

    def __rfloordiv__(self, other):
        pass

    def __mod__(self, other):
        pass

    def __rmod__(self, other):
        pass

    def __and__(self, other):
        pass

    def __rand__(self, other):
        pass

    def __or__(self, other):
        pass

    def __ror__(self, other):
        pass

    def __xor__(self, other):
        pass

    def __rxor__(self, other):
        pass

    def __lshift__(self, other):
        pass

    def __rlshift__(self, other):
        pass

    def __rshift__(self, other):
        pass

    def __rrshift__(self, other):
        pass


def int8(v) -> Int8:
    """Shorthand of `Int8(v)`."""
    return Int8(v)


def int16(v) -> Int16:
    """Shorthand of `Int16(v)`."""
    return Int16(v)


def int32(v) -> Int32:
    """Shorthand of `Int32(v)`."""
    return Int32(v)


def int64(v) -> Int64:
    """Shorthand of `Int64(v)`."""
    return Int64(v)


def uint8(v) -> UInt8:
    """Shorthand of `UInt8(v)`."""
    return UInt8(v)


def uint16(v) -> UInt16:
    """Shorthand of `UInt16(v)`."""
    return UInt16(v)


def uint32(v) -> UInt32:
    """Shorthand of `UInt32(v)`."""
    return UInt32(v)


def uint64(v) -> UInt64:
    """Shorthand of `UInt64(v)`."""
    return UInt64(v)
